__version__ = "0.0.1"

from .ib_protocol import IBProtocol
from .ib_factory import IBClientFactory
from .xmlrpc_server import XMLRPCServer
from .exceptions import IBException
from .applications import Solid



"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""


"""
Collection of misc tools
"""


import sys
import logging
import inspect

from .common import UNSET_INTEGER, UNSET_DOUBLE, UNSET_LONG


logger = logging.getLogger(__name__)


# I use this just to visually emphasize it's a wrapper overriden method
def iswrapper(fn):
    return fn


class BadMessage(Exception):
    def __init__(self, text):
        self.text = text


class LogFunction(object):
    def __init__(self, text, logLevel):
        self.text = text
        self.logLevel = logLevel

    def __call__(self, fn):
        def newFn(origSelf, *args, **kwargs):
            if logger.getLogger().isEnabledFor(self.logLevel):
                argNames = [argName for argName in inspect.getfullargspec(fn)[0] if argName != 'self']
                logger.log(self.logLevel,
                    "{} {} {} kw:{}".format(self.text, fn.__name__,
                        [nameNarg for nameNarg in zip(argNames, args) if nameNarg[1] is not origSelf], kwargs))
            fn(origSelf, *args)
        return newFn


def current_fn_name(parent_idx = 0):
    #depth is 1 bc this is already a fn, so we need the caller
    return sys._getframe(1 + parent_idx).f_code.co_name


def setattr_log(self, var_name, var_value):
    #import code; code.interact(local=locals())
    logger.debug("%s %s %s=|%s|", self.__class__, id(self), var_name, var_value)
    super(self.__class__, self).__setattr__(var_name, var_value)


SHOW_UNSET = True
def decode(the_type, fields, show_unset = False):
    try:
        s = next(fields)
    except StopIteration:
        raise BadMessage("no more fields")

    logger.debug("decode %s %s", the_type, s)

    if the_type is str:
        if type(s) is str:
            return s
        elif type(s) is bytes:
            return s.decode(errors='backslashreplace')
        else:
            raise TypeError("unsupported incoming type " + type(s) + " for desired type 'str")

    orig_type = the_type
    if the_type is bool:
        the_type = int

    if show_unset:
        if s is None or len(s) == 0:
            if the_type is float:
                n = UNSET_DOUBLE
            elif the_type is int:
                n = UNSET_INTEGER
            else:
                raise TypeError("unsupported desired type for empty value" + the_type)
        else:
            n = the_type(s)
    else:
        n = the_type(s or 0)

    if orig_type is bool:
        n = False if n == 0 else True

    return n



def ExerciseStaticMethods(klass):

    import types
    #import code; code.interact(local=dict(globals(), **locals()))
    for (_, var) in inspect.getmembers(klass):
        #print(name, var, type(var))
        if type(var) == types.FunctionType:
            print("Exercising: %s:" % var)
            print(var())
            print()

def floatToStr(val):
    return str(val) if val != UNSET_DOUBLE else "";

def longToStr(val):
    return str(val) if val != UNSET_LONG else "";





import struct
from typing import Optional

from loguru import logger
from twisted.internet.protocol import connectionDone
from twisted.protocols.basic import Int32StringReceiver
from twisted.protocols.policies import TimeoutMixin

from sibi.ibapi.server_versions import MIN_CLIENT_VER, MAX_CLIENT_VER

PROTOCOL_TIMEOUT = 10
(DISCONNECTED, CONNECTING, CONNECTED, REDIRECT) = range(4)


class IBProtocol(Int32StringReceiver, TimeoutMixin):

    def __init__(self):
        self.setTimeout(PROTOCOL_TIMEOUT)

    def sendMsg(self, msg):
        self.setTimeout(PROTOCOL_TIMEOUT)
        self.transport.write(msg)

    def stringReceived(self, text):
        self.setTimeout(None)
        fields = self.split_fields(text)

        if self.factory.connState == CONNECTING:
            if len(fields) == 2:
                logger.debug(f"Consibiating connection for {text}")
                self.finalize_connection(fields)
            else:
                logger.error(f"Error while connecting")
            return
        else:
            self.factory.decoder.interpret(fields)

    def connectionLost(self, reason=connectionDone):
        """
        This just sets the IB EClient connection state when disconnected.
        """

        logger.warning("Closing connection")
        if self.factory.connState == CONNECTING:
            logger.warning("Reconnecting")

        self.factory.setConnState(DISCONNECTED)
        logger.warning(reason.getErrorMessage())

    def isConnected(self):
        """
        Determines if we are connected to TWS
        """
        return self.factory.connState == CONNECTED

    def connectionMade(self):
        """
        As soon as connection is made, we need to send a message containing API version to TWS.
        This is the first part of IB's connection procedure.
        """
        logger.info("Connecting to TWS")
        self.factory.setConnState(CONNECTING)
        self.factory.conn = self
        v100prefix = "API\0"
        v100version = "v%d..%d" % (MIN_CLIENT_VER, MAX_CLIENT_VER)
        msg = self.make_msg(v100version)

        msg2 = str.encode(v100prefix, 'ascii') + msg

        self.sendMsg(msg2)

    def finalize_connection(self, fields):
        """
        Finalizes connection with TWS
        """
        (server_version, conn_time) = fields
        server_version = int(server_version)
        logger.info(f"{self.factory.name}: ANSWER Version:{server_version} time:{conn_time}")
        self.factory.connTime = conn_time
        self.factory.serverVersion_ = server_version
        self.factory.decoder.serverVersion = self.factory.serverVersion()

        logger.debug(f"{self.factory.name}: sent startApi")
        self.factory.setConnState(CONNECTED)
        self.factory.startApi()

    @staticmethod
    def split_fields(buffer: bytes, separator: Optional[bytes] = b"\0") -> tuple:
        """
        Payload is made of fields terminated/separated by NULL chars.
        We split the bytes using the null b"\0" separator and return a tuple containing fields.
        """

        fields = buffer.split(separator)

        return tuple(fields[:-1])  # Last is always empty

    @staticmethod
    def make_field(value: str) -> str:
        """
        IB detects different fields by splitting a string using a null separator (\0)
        This just adds the NULL string terminator to a common string
        """

        if value is None:
            raise ValueError("Cannot send None to TWS")

        # bool type is encoded as int
        if type(value) is bool:
            value = int(value)

        # we add the null terminator and return
        field = str(value) + '\0'
        return field

    @staticmethod
    def make_msg(text) -> bytes:
        """ adds the length prefix """
        msg = struct.pack("!I%ds" % len(text), len(text), str.encode(text))
        return msg



class IBException(Exception):
    def __init__(self, code, message):
        self.code = code
        self.message = message

    def __repr__(self):
        return self.__dict__

    def __str__(self):
        return self.__dict__



import sys

from loguru import logger


WEBHOOK_URL = 'http://localhost:8000/hooks/{channel}'

fmt = '<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: ^8}</level> | ' \
      '<cyan>{name: <20}</cyan> | <cyan>{function: <20}</cyan> | ' \
      '<cyan>{line: ^4}</cyan> | <level>{message}</level>'

logger.configure(
    handlers=[
        {"sink": sys.stdout, "format": fmt, "level": "INFO"},
    ],
)



"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""

class Object(object):

    def __str__(self):
        return "Object"

    def __repr__(self):
        return str(id(self)) + ": " + self.__str__()

  



import json
from functools import wraps

import redis
import treq
from loguru import logger
from twisted.internet import defer

from sibi.settings import WEBHOOK_URL

try:
    r = redis.StrictRedis(host="localhost", port=6379, db=0)
except Exception as e:
    logger.error(e)


def request(method=None, order=False):
    def _decorate(function):
        @wraps(function)
        def wrapper(self, *args, **kwargs):
            logger.info(f"Start processing {function.__name__}")
            if order:
                identifier = self.getNextOrderId()
                deferredDict = self.deferredOrdersRequests
            else:
                identifier = self.getNextReqId()
                deferredDict = self.deferredRequests
            deferredDict[identifier] = defer.Deferred()
            deferredDict[identifier].addCallbacks(json.dumps, self.error)
            _ = function(self, identifier, *args, **kwargs)
            return _

        return wrapper

    if method:
        return _decorate(method)
    return _decorate


def resolve(method=None, order=False):
    def _decorate(function):
        @wraps(function)
        def wrapper(self, *args, **kwargs):
            if order:
                deferredDict = self.deferredOrdersRequests
                deferredResultsDict = self.deferredOrdersResults
            else:
                deferredDict = self.deferredRequests
                deferredResultsDict = self.deferredResults
            _ = function(self, *args, **kwargs)
            identifier = args[0]
            deferredDict[identifier].callback(deferredResultsDict[identifier])
            del deferredDict[identifier]
            del deferredResultsDict[identifier]
            logger.info(f"Finish processing {function.__name__}")
            return _

        return wrapper

    if method:
        return _decorate(method)
    return _decorate


def publish(method=None, order=False):
    def _decorate(function):
        @wraps(function)
        def wrapper(self, *args, **kwargs):
            data = function(self, *args, **kwargs)
            if order:
                additionalInfoDict = self.additionalOrderInfo
            else:
                additionalInfoDict = self.additionalRequestInfo
            identifier = args[0]
            if additionalInfoDict.get(identifier):
                data = {**data, **additionalInfoDict[identifier]}
            d = treq.get(url=WEBHOOK_URL.format(channel=function.__name__), params=data)
            d.addCallbacks(logger.warning, logger.error)
            data = json.dumps(data)
            r.publish(function.__name__, data)
            logger.info(f"Pushing {data} on channel {function.__name__}")
            return data

        return wrapper

    if method:
        return _decorate(method)
    return _decorate


def append(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        data = method(self, *args, **kwargs)
        reqId = args[0]
        if reqId not in self.deferredResults.keys():
            self.deferredResults[reqId] = []
        self.deferredResults[reqId].append(data)

        logger.debug(f"Start publishing {reqId} on channel {method.__name__}")
        return data

    return wrapper



"""
Copyright (C) 2019 Interactive Brokers LLC. All rights reserved. This code is subject to the terms
 and conditions of the IB API Non-Commercial License or the IB API Commercial License, as applicable.
"""

import sys

from .enum_implem import Enum
from .object_implem import Object


NO_VALID_ID = -1
MAX_MSG_LEN = 0xFFFFFF # 16Mb - 1byte

UNSET_INTEGER = 2 ** 31 - 1
UNSET_DOUBLE = sys.float_info.max
UNSET_LONG = 2 ** 63 - 1

TickerId = int
OrderId  = int
TagValueList = list

FaDataType = int
FaDataTypeEnum = Enum("N/A", "GROUPS", "PROFILES", "ALIASES")

MarketDataType = int
MarketDataTypeEnum = Enum("N/A", "REALTIME", "FROZEN", "DELAYED", "DELAYED_FROZEN")

Liquidities = int
LiquiditiesEnum = Enum("None", "Added", "Remove", "RoudedOut")

SetOfString = set
SetOfFloat = set
ListOfOrder = list
ListOfFamilyCode = list
ListOfContractDescription = list
ListOfDepthExchanges = list
ListOfNewsProviders = list
SmartComponentMap = dict
HistogramDataList = list
ListOfPriceIncrements = list
ListOfHistoricalTick = list
ListOfHistoricalTickBidAsk = list
ListOfHistoricalTickLast = list

class BarData(Object):
    def __init__(self):
        self.date = ""
        self.open = 0.
        self.high = 0.
        self.low = 0.
        self.close = 0.
        self.volume = 0
        self.barCount = 0
        self.average = 0.

    def __str__(self):
        return "Date: %s, Open: %f, High: %f, Low: %f, Close: %f, Volume: %d, Average: %f, BarCount: %d" % (self.date, self.open, self.high,
            self.low, self.close, self.volume, self.average, self.barCount)

class RealTimeBar(Object):
    def __init__(self, time = 0, endTime = -1, open_ = 0., high = 0., low = 0., close = 0., volume = 0., wap = 0., count = 0):
        self.time = time
        self.endTime = endTime
        self.open_ = open_
        self.high = high
        self.low = low
        self.close = close
        self.volume = volume
        self.wap = wap
        self.count = count

    def __str__(self):
        return "Time: %d, Open: %f, High: %f, Low: %f, Close: %f, Volume: %d, WAP: %f, Count: %d" % (self.time, self.open_, self.high,
            self.low, self.close, self.volume, self.wap, self.count)

class HistogramData(Object):
    def __init__(self):
        self.price = 0.
        self.count = 0

    def __str__(self):
        return "Price: %f, Count: %d" % (self.price, self.count)

class NewsProvider(Object):
    def __init__(self):
        self.code = ""
        self.name = ""

    def __str__(self):
        return "Code: %s, Name: %s" % (self.code, self.name)

class DepthMktDataDescription(Object):
    def __init__(self):
        self.exchange = ""
        self.secType = ""
        self.listingExch = ""
        self.serviceDataType = ""
        self.aggGroup = UNSET_INTEGER

    def __str__(self):
        if (self.aggGroup!= UNSET_INTEGER):
            aggGroup = self.aggGroup
        else:
            aggGroup = ""
        return "Exchange: %s, SecType: %s, ListingExchange: %s, ServiceDataType: %s, AggGroup: %s, " % (self.exchange, self.secType, self.listingExch,self.serviceDataType, aggGroup)

class SmartComponent(Object):
    def __init__(self):
        self.bitNumber = 0
        self.exchange = ""
        self.exchangeLetter = ""

    def __str__(self):
        return "BitNumber: %d, Exchange: %s, ExchangeLetter: %s" % (self.bitNumber, self.exchange, self.exchangeLetter)

class TickAttrib(Object):
    def __init__(self):
        self.canAutoExecute = False
        self.pastLimit = False
        self.preOpen = False

    def __str__(self):
        return "CanAutoExecute: %d, PastLimit: %d, PreOpen: %d" % (self.canAutoExecute, self.pastLimit, self.preOpen)

class TickAttribBidAsk(Object):
    def __init__(self):
        self.bidPastLow = False
        self.askPastHigh = False

    def __str__(self):
        return "BidPastLow: %d, AskPastHigh: %d" % (self.bidPastLow, self.askPastHigh)

class TickAttribLast(Object):
    def __init__(self):
        self.pastLimit = False
        self.unreported = False

    def __str__(self):
        return "PastLimit: %d, Unreported: %d" % (self.pastLimit, self.unreported)

class FamilyCode(Object):
    def __init__(self):
        self.accountID = ""
        self.familyCodeStr = ""

    def __str__(self):
        return "AccountId: %s, FamilyCodeStr: %s" % (self.accountID, self.familyCodeStr)

class PriceIncrement(Object):
    def __init__(self):
        self.lowEdge = 0.
        self.increment = 0.

    def __str__(self):
        return "LowEdge: %f, Increment: %f" % (self.lowEdge, self.increment)

class HistoricalTick(Object):
    def __init__(self):
        self.time = 0
        self.price = 0.
        self.size = 0

    def __str__(self):
        return "Time: %d, Price: %f, Size: %d" % (self.time, self.price, self.size)

class HistoricalTickBidAsk(Object):
    def __init__(self):
        self.time = 0
        self.tickAttribBidAsk = TickAttribBidAsk()
        self.priceBid = 0.
        self.priceAsk = 0.
        self.sizeBid = 0
        self.sizeAsk = 0

    def __str__(self):
        return "Time: %d, TickAttriBidAsk: %s, PriceBid: %f, PriceAsk: %f, SizeBid: %d, SizeAsk: %d" % (self.time, self.tickAttribBidAsk, self.priceBid, self.priceAsk, self.sizeBid, self.sizeAsk)

class HistoricalTickLast(Object):
    def __init__(self):
        self.time = 0
        self.tickAttribLast = TickAttribLast()
        self.price = 0.
        self.size = 0
        self.exchange = ""
        self.specialConditions = ""

    def __str__(self):
        return "Time: %d, TickAttribLast: %s, Price: %f, Size: %d, Exchange: %s, SpecialConditions: %s" % (self.time, self.tickAttribLast, self.price, self.size, self.exchange, self.specialConditions)





import datetime
import sys
from typing import List

from loguru import logger
from twisted.internet import reactor
from twisted.web import xmlrpc
from twisted.web.client import HTTPConnectionPool, Agent

from sibi.ibapi.common import BarData
from sibi.ibapi.contract import Contract, ComboLeg
from sibi.ibapi.order import Order


class XMLRPCServer(xmlrpc.XMLRPC):

    def __init__(self, factory):
        super(XMLRPCServer, self).__init__()
        pool = HTTPConnectionPool(reactor, persistent=True)
        pool.maxPersistentPerHost = 10
        self.agent = Agent(reactor, pool=pool)
        self.factory = factory

        xmlrpc.addIntrospection(self)

    def xmlrpc_reqContractDetails(
            self, symbol: str = '', secType: str = 'STK', currency: str = 'USD', exchange: str = 'SMART',
            lastTradeDateOrContractMonth: str = '', strike: str = '', right: str = ''):
        """ This returns information about a contract's conID, symbol, local symbol, currency, etc.
         reqContractDetails takes as an argument a Contract object which may uniquely match one contract,
         and unlike other API functions it can also take a Contract object which matches multiple contracts
         in IB's database. When there are multiple matches, a list is returned

        Args:
            symbol (str): IB symbol (eg. SPY, AAPL, DAX)
            secType (str): The type of security (eg. IND, STK, OPT)
            currency (str): The currency of the security (eg. EUR, USD)
            exchange (str): The exchange (eg SMART, CBOE)
            lastTradeDateOrContractMonth (str): This is a date for OPTIONS (eg. 20210104) or a month for FUTURES (eg. 202103)
            strike (str): Strike price for options
            right (str): Right for options (eg. C or CALL, P or PUT)
        """
        contract = Contract()
        contract.symbol = symbol  # "EUR"
        contract.secType = secType  # "CASH"
        contract.currency = currency  # "GBP"
        contract.exchange = exchange  # "IDEALPRO"
        contract.lastTradeDateOrContractMonth = lastTradeDateOrContractMonth
        contract.strike = strike
        contract.right = right
        contract.includeExpired = False
        result = self.factory.reqContractDetails(contract)

        return result

    def xmlrpc_reqHistoricalData(
            self, symbol: str = '', secType: str = '', currency: str = '', exchange: str = '', durationStr: str = '1 M',
            barSizeSetting: str = '1 hour', whatToShow: str = 'MIDPOINT', useRTH: int = 1, formatDate: int = 1,
            keepUpToDate: bool = False, chartOptions=None) -> List[BarData]:
        if chartOptions is None:
            chartOptions = []
        contract = Contract()
        contract.symbol = symbol  # "EUR"
        contract.secType = secType  # "CASH"
        contract.currency = currency  # "GBP"
        contract.exchange = exchange  # "IDEALPRO"
        query_time = (datetime.datetime.today() - datetime.timedelta(days=5)).strftime("%Y%m%d %H:%M:%S")
        result = self.factory.reqHistoricalData(
            contract, query_time, durationStr, barSizeSetting, whatToShow, useRTH, formatDate, keepUpToDate,
            chartOptions)
        return result

    def xmlrpc_reqMktData(
            self, conId: str = 0, symbol: str = '', secType: str = '', currency: str = '', exchange: str = '',
            localSymbol: str = ''):
        contract = Contract()
        contract.conId = conId
        contract.symbol = symbol  # "EUR"
        contract.secType = secType  # "CASH"
        contract.currency = currency  # "GBP"
        contract.exchange = exchange  # "IDEALPRO"
        contract.localSymbol = localSymbol
        contract.includeExpired = False
        result = self.factory.reqMktData(contract)
        return result

    def xmlrpc_cancelMktData(self, tickerId: int = -1):
        """ Cancels market data subscriptions.

            Args:
                tickerId (int): The ticker ID which was specified in original market data request.
                    Cancelling a subscription allows the user to make a subscription to a different contract and remain
                    within the level 1 market data lines allowance.
        """
        result = self.factory.cancelMktData(tickerId)
        return result

    def xmlrpc_placeOrder(
            self, symbol: str = '', secType: str = 'STK', currency: str = 'USD', exchange: str = 'SMART',
            lastTradeDateOrContractMonth: str = '', strike: str = '', right: str = '', comboLegs=None,
            action: str = 'BUY', orderType: str = 'MIDPRICE', totalQuantity: int = 1,
            limitPrice: float = sys.float_info.max):

        contract = Contract()
        contract.symbol = symbol
        contract.secType = secType
        contract.currency = currency
        contract.exchange = exchange
        contract.lastTradeDateOrContractMonth = lastTradeDateOrContractMonth
        contract.strike = strike
        contract.right = right
        contract.comboLegs = []

        if comboLegs and secType == 'BAG':
            for comboLeg in comboLegs:
                leg = ComboLeg()
                for key, value in comboLeg.items():
                    setattr(leg, key, value)
                contract.comboLegs.append(leg)

        order = Order()
        order.action = action
        order.orderType = orderType
        order.totalQuantity = totalQuantity
        order.lmtPrice = limitPrice
        logger.error(contract)
        logger.warning(order)
        return self.factory.getNextOrderId()



from twisted.internet import reactor, endpoints
from twisted.protocols.policies import ThrottlingFactory
from twisted.web import server

from sibi.ib_factory import IBClientFactory
from sibi.xmlrpc_server import XMLRPCServer
from sibi.settings import logger


class Solid:

    def __init__(self,
                 client_id: int = 0,
                 tws_host: str = '192.168.74.130',
                 tws_port: int = 7498,
                 tws_max_connections: int = 49,
                 xmlrpc_port: int = 7080):
        logger.info("Creating APP")
        ib_factory = IBClientFactory(client_id)
        reactor.connectTCP(tws_host, tws_port, ib_factory)

        # create XMLRPC SERVER ENDPOINT
        xmlrpc_resource = XMLRPCServer(ib_factory)
        xmlrpc_factory = ThrottlingFactory(server.Site(xmlrpc_resource), tws_max_connections)

        endpoint = endpoints.TCP4ServerEndpoint(reactor, xmlrpc_port)
        endpoint.listen(xmlrpc_factory)

    def run(self):
        reactor.run()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sibi', 'sibi.ibapi']

package_data = \
{'': ['*']}

install_requires = \
['Twisted>=20.3.0,<21.0.0',
 'loguru>=0.5.3,<0.6.0',
 'redis>=3.5.3,<4.0.0',
 'treq>=20.9.0,<21.0.0',
 'typer>=0.3.2,<0.4.0']

setup_kwargs = {
    'name': 'sibi',
    'version': '0.1.0',
    'description': 'Synchronous Interactiver Brokers Interface',
    'long_description': None,
    'author': 'Manuel Fedele',
    'author_email': 'manuelfedele@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from loguru import logger
from twisted.internet.defer import Deferred
from twisted.internet.protocol import ReconnectingClientFactory
from twisted.internet.tcp import Connector
from twisted.python.failure import Failure

from sibi.ibapi import decoder
from sibi.ibapi.client import EClient
from sibi.ibapi.common import TickerId, BarData, TagValueList, TickAttrib, OrderId
from sibi.ibapi.contract import Contract, ContractDetails
from sibi.ibapi.order import Order
from sibi.ibapi.order_state import OrderState
from sibi.ibapi.ticktype import TickType, TickTypeEnum
from sibi.ibapi.wrapper import EWrapper
from sibi.decorators import append, publish, request, resolve
from sibi.exceptions import IBException
from sibi.ib_protocol import IBProtocol


class IBClientFactory(ReconnectingClientFactory, EClient, EWrapper):
    """
    This is the Factory that instantiates the IBProtocols instance.

    Args:
        clientId (int): The clientId for TWS
    """
    protocol = IBProtocol

    def __init__(self, client_id: int) -> None:
        EClient.__init__(self, wrapper=self)
        self.name = "IBClientFactory"
        self.decoder = decoder.Decoder(self, self.serverVersion())
        self.clientId = client_id
        self.currentReqId = 1
        self.nextValidOrderId = -1
        self.deferredRequests = {}
        self.deferredResults = {}
        self.additionalRequestInfo = {}

        self.deferredOrdersRequests = {}
        self.deferredOrdersResults = {}
        self.additionalOrderInfo = {}
        self.activeMktDataSubscriptions = []

    def clientConnectionLost(self, connector: Connector, reason: Failure) -> None:
        """ Internal reconnection method in case of connection lose """
        logger.warning(f"Lost connection.  Reason: {reason.getErrorMessage()}")
        ReconnectingClientFactory.clientConnectionLost(self, connector, reason)

    def clientConnectionFailed(self, connector: Connector, reason: Failure) -> None:
        """ Internal reconnection method in case of connection fail """
        logger.warning(f"Connection failed.  Reason: {reason.getErrorMessage()}")
        ReconnectingClientFactory.clientConnectionFailed(self, connector, reason)

    def getNextReqId(self) -> int:
        """ Increments the reqId for TWS. This is automatically called by **@request** decorator """

        reqId = self.currentReqId
        self.currentReqId = self.currentReqId + 1
        return reqId

    def getNextOrderId(self, numIds: int = -1):
        super(IBClientFactory, self).reqIds(numIds)
        return self.nextValidOrderId

    def nextValidId(self, orderId: int) -> int:
        """
        nextValidId event provides the next valid identifier needed to place an order.
        This is automatically called by TWS
        """

        super(IBClientFactory, self).nextValidId(orderId)
        self.nextValidOrderId = orderId
        return orderId

    @request(order=True)
    def placeOrder(self, orderId: OrderId, contract: Contract, order: Order):
        """
        Places an order.
        Immediately after the order is submitted correctly, the TWS will start sending events concerning
        the order's activity via **openOrder** and **orderStatus** methods
        """
        super(IBClientFactory, self).placeOrder(orderId, contract, order)
        return self.deferredRequests[orderId]

    @resolve(order=True)
    def openOrder(self, orderId: OrderId, contract: Contract, order: Order, orderState: OrderState):
        """ TODO: Test this """
        pass

    @request
    def reqContractDetails(self, reqId: int, contract: Contract, **kwargs) -> Deferred:
        super(IBClientFactory, self).reqContractDetails(reqId, contract)
        return self.deferredRequests[reqId]

    @request
    def reqHistoricalData(
            self, reqId: int, contract: Contract, endDateTime: str, durationStr: str, barSizeSetting: str,
            whatToShow: str, useRTH: int, formatDate: int, keepUpToDate: bool, chartOptions: TagValueList,
            **kwargs) -> Deferred:
        """
        Requests historical data for contract.
        """

        super(IBClientFactory, self).reqHistoricalData(
            reqId, contract, endDateTime, durationStr, barSizeSetting, whatToShow, useRTH, formatDate, keepUpToDate,
            chartOptions)

        return self.deferredRequests[reqId]

    @request
    @resolve
    def reqMktData(
            self, reqId: int, contract: Contract, genericTickList: str = "", snapshot: bool = False,
            regulatorySnapshot: bool = False, mktDataOptions=None, **kwargs, ) -> None:

        """ Starts live market data polling for specified contract """
        if mktDataOptions is None:
            mktDataOptions = []

        super(IBClientFactory, self).reqMktData(
            reqId, contract, genericTickList, snapshot, regulatorySnapshot, mktDataOptions)
        self.deferredResults[reqId] = {"reqId": reqId}
        self.activeMktDataSubscriptions.append(reqId)
        self.additionalRequestInfo[reqId] = contract.__dict__
        return self.deferredRequests[reqId]

    @request
    @resolve
    def cancelMktData(self, reqId: int, tickerId: TickerId):
        canceledIDs = []
        if tickerId == -1:
            for _tickerId in self.activeMktDataSubscriptions:
                super(IBClientFactory, self).cancelMktData(int(_tickerId))
                canceledIDs.append(_tickerId)
            self.activeMktDataSubscriptions = []
            self.deferredResults[reqId] = {"reqId": canceledIDs}
        else:
            super(IBClientFactory, self).cancelMktData(tickerId)
            self.deferredResults[reqId] = {"reqId": reqId}
        return self.deferredRequests[reqId]

    @append
    def contractDetails(self, reqId: int, contractDetails: ContractDetails) -> dict:
        return contractDetails.contract.__dict__

    @append
    def historicalData(self, reqId: int, bar: BarData) -> dict:
        """ Adds a BAR to reqId deferred """
        return bar.__dict__

    @publish
    def tickPrice(self, reqId: int, tickType: TickType, price: float, attrib: TickAttrib) -> dict:
        """ This method is called by IB server when a new tickPrice is available for active live data market lines """

        data = {
            "reqId": reqId,
            "tickType": TickTypeEnum.to_str(tickType),
            "price": price,
        }

        return data

    def tickSize(self, reqId: int, tickType: TickType, size: int) -> dict:
        data = {
            "reqId": reqId,
            "tickType": TickTypeEnum.to_str(tickType),
            "size": size,
        }
        return data

    def tickString(self, reqId: int, tickType: TickType, value: str) -> dict:
        data = {
            "reqId": reqId,
            "tickType": TickTypeEnum.to_str(tickType),
            "value": value,
        }
        return data

    def tickGeneric(self, reqId: int, tickType: TickType, value: float) -> dict:
        data = {
            "reqId": reqId,
            "tickType": TickTypeEnum.to_str(tickType),
            "value": value,
        }
        return data

    @publish
    def historicalDataUpdate(self, reqId: int, bar: BarData) -> dict:
        data = {
            "reqId": reqId,
            "barData": BarData.__dict__
        }
        return data

    @resolve
    def historicalDataEnd(self, reqId: int, start: str, end: str) -> None:
        pass

    @resolve
    def contractDetailsEnd(self, reqId: int) -> None:
        pass

    def error(self, reqId: TickerId, errorCode: int = -1, errorString: str = "GenericError") -> None:
        NON_ERROR_CODES = [2104, 2106, 2158]
        if errorCode in NON_ERROR_CODES:
            logger.info(errorString)
        else:
            logger.error(f"Error. Id: {reqId} Code: {errorCode} Msg: {errorString}")

        if reqId in self.deferredRequests.keys():
            self.deferredRequests[reqId].callback(IBException(errorCode, errorString).__dict__)
            del self.deferredRequests[reqId]
